import 'dart:convert';
import 'package:dart_icalendar/models/extensions/duration_extension.dart';

class Alarm {
  Duration duration;
  String? description;

  Alarm({
    required this.duration,
    this.description,
  });

  String toICalString() {
    final buffer = StringBuffer();
    buffer.writeln('BEGIN:VALARM');
    buffer.writeln('TRIGGER:-PT${duration.inMinutes}M');
    buffer.writeln('ACTION:DISPLAY');
    if (description != null) {
      buffer.writeln('DESCRIPTION:$description');
    }
    buffer.writeln('END:VALARM');
    return buffer.toString();
  }

  factory Alarm.fromICalString(String icalString) {
    final durationRegex = RegExp(
      r'^TRIGGER:(?<sign>-)?P(?:(?<days>\d+)D)?T?(?:(?<hours>\d+)H)?(?:(?<minutes>\d+)M)?(?:(?<seconds>\d+)S)?|TRIGGER:(?<sign2>-)?PT(?:(?<hours2>\d+)H)?(?:(?<minutes2>\d+)M)?(?:(?<seconds2>\d+)S)?$',
      caseSensitive: false,
      multiLine: true,
    );

    final durationMatch = durationRegex.firstMatch(icalString);
    if (durationMatch == null) {
      throw FormatException('Invalid duration format');
    }

    final isNegative = durationMatch.namedGroup('sign') != null ||
        durationMatch.namedGroup('sign2') != null;
    final days = int.tryParse(durationMatch.namedGroup('days') ?? '0') ?? 0;
    final hours = int.tryParse(durationMatch.namedGroup('hours') ?? '0') ??
        int.tryParse(durationMatch.namedGroup('hours2') ?? '0') ??
        0;
    final minutes = int.tryParse(durationMatch.namedGroup('minutes') ?? '0') ??
        int.tryParse(durationMatch.namedGroup('minutes2') ?? '0') ??
        0;
    final seconds = int.tryParse(durationMatch.namedGroup('seconds') ?? '0') ??
        int.tryParse(durationMatch.namedGroup('seconds2') ?? '0') ??
        0;

    var duration = Duration(
      days: days,
      hours: hours,
      minutes: minutes,
      seconds: seconds,
    );

    duration = duration.multiplyWithSign(duration, isNegative ? 1 : -1);

    final descriptionMatch = RegExp(r'DESCRIPTION:(.*)').firstMatch(icalString);
    final description = descriptionMatch?.group(1)?.trim();

    return Alarm(duration: duration, description: description);
  }

  Alarm copyWith({
    Duration? duration,
    String? description,
  }) {
    return Alarm(
      duration: duration ?? this.duration,
      description: description ?? this.description,
    );
  }

  Map<String, dynamic> toMap() {
    return {
      'duration': duration.inMilliseconds,
      'description': description,
    };
  }

  factory Alarm.fromMap(Map<String, dynamic> map) {
    return Alarm(
      duration: Duration(milliseconds: map['duration']),
      description: map['description'],
    );
  }

  String toJson() => json.encode(toMap());

  factory Alarm.fromJson(String source) => Alarm.fromMap(json.decode(source));

  @override
  String toString() => 'Alarm(duration: $duration, description: $description)';

  @override
  bool operator ==(Object other) {
    if (identical(this, other)) return true;

    return other is Alarm &&
        other.duration == duration &&
        other.description == description;
  }

  @override
  int get hashCode => duration.hashCode ^ description.hashCode;
}
