import 'dart:convert';

import 'package:collection/collection.dart';
import 'package:intl/intl.dart';

import 'package:dart_icalendar/models/attachment_model.dart';
import 'package:dart_icalendar/models/location_model.dart';
import 'package:dart_icalendar/models/person_model.dart';
import 'package:dart_icalendar/models/recurrence_model.dart';
import 'package:dart_icalendar/models/status_model.dart';
import 'package:dart_icalendar/models/time_zone_model.dart';
import 'package:dart_icalendar/models/x_date_model.dart';

class Event {
  String uid;
  String? summary;
  String? description;
  List<Attachment>? attachments;
  List<Attendee>? attendees;
  Organizer? organizer;
  DateTime? stamp;
  DateTime? created;
  DateTime? lastModified;
  DateTime? startDate;
  DateTime? endDate;
  int? sequence;
  String? transparency;
  EventStatus? status;
  TimeZone? timeZone;
  RecurrenceRule? recurrenceRule;
  List<XDate>? xDates;
  Location? location;
  String? recurrenceId;
  String? eventClass;
  Map<String, String>? customProperties;
  List<Attachment>? attachFiles;
  Availability? get availability {
    for (String key in customProperties?.keys ?? []) {
      if (key == 'X-AVAILABILITY') {
        return Availability.values[int.tryParse(customProperties![key]!) ?? 0];
      }
    }
    return null;
  }

  Event({
    required this.uid,
    this.summary,
    this.description,
    this.attachments,
    this.attendees,
    this.organizer,
    this.stamp,
    this.created,
    this.lastModified,
    this.startDate,
    this.endDate,
    this.sequence,
    this.transparency,
    this.status,
    this.timeZone,
    this.recurrenceRule,
    this.xDates,
    this.location,
    this.recurrenceId,
    this.eventClass,
    this.customProperties,
    this.attachFiles,
  });

  DateTime? get start {
    return startDate;
  }

  DateTime? get end {
    return endDate;
  }

  DateTime? get createdDate {
    return created!;
  }

  DateTime? get lastModifiedDate {
    return lastModified!;
  }

  DateTime? get stampDate {
    return stamp!;
  }

  TimeZone? get timeZoneValue {
    return timeZone!;
  }

  set availability(Availability? value) {
    customProperties ??= {};
    if (value == null) {
      customProperties?.remove('X-AVAILABILITY');
      return;
    }
    customProperties?['X-AVAILABILITY'] = value.index.toString();
  }

  List<Event>? getRecurrenceInstancesForTimeFrame(
      DateTime periodStart, DateTime periodEnd) {
    assert(periodStart.isBefore(periodEnd));

    DateTime defaultStartDate = start!;

    List<DateTime> recurrenceDates = recurrenceRule!.getRecurrencesForPeriod(
        periodStart, periodEnd, defaultStartDate, end!);

    List<Event> events = [];
    for (DateTime recurrenceDate in recurrenceDates) {
      if (recurrenceDate.isAfter(periodEnd)) {
        break;
      }
      if (recurrenceDate.isAfter(periodStart) ||
          recurrenceDate == periodStart) {
        events.add(Event(
          uid: uid,
          summary: summary,
          description: description,
          attachments: attachments,
          attendees: attendees,
          organizer: organizer,
          stamp: stamp,
          created: created,
          lastModified: lastModified,
          startDate: recurrenceDate,
          endDate: recurrenceDate.add(end!.difference(start!)),
          sequence: sequence,
          transparency: transparency,
          status: status,
          timeZone: timeZone,
          recurrenceRule: recurrenceRule,
          xDates: xDates,
          location: location,
          recurrenceId: null,
          eventClass: eventClass,
          customProperties: customProperties,
          attachFiles: attachFiles,
        ));
      }
    }
    return events;
  }

  String _foldLine(String line) {
    final maxLineLength = 75;
    final sb = StringBuffer();
    line = line.replaceAll('\r', '');
    while (line.length > maxLineLength) {
      sb.write('${line.substring(0, maxLineLength)}\r\n ');
      line = line.substring(maxLineLength);
    }
    sb.write('$line\r');
    return sb.toString();
  }

  String toICalString() {
    final buffer = StringBuffer();
    buffer.writeln(_foldLine('BEGIN:VEVENT'));

    if (startDate != null) {
      if (startDate!.isAfter(DateTime(2022, 10, 09)) &&
          startDate!.isBefore(DateTime(2022, 10, 11))) {
        print('startDate: $startDate');
        print('endDate: $endDate');
        print('timeZone: $timeZone');
      }
      if (timeZone != null) {
        if (timeZone!.tzId == 'DATE') {
          buffer.writeln(_foldLine(
              'DTSTART;VALUE=DATE:${DateFormat('yyyyMMdd').format(startDate!)}'));
          if (endDate != null) {
            buffer.writeln(_foldLine(
                'DTEND;VALUE=DATE:${DateFormat('yyyyMMdd').format(endDate!)}'));
          }
        } else {
          buffer.writeln(_foldLine(
              'DTSTART;TZID=${timeZone!.tzId}:${DateFormat('yyyyMMdd\'T\'HHmmss').format(startDate!)}'));
          if (endDate != null) {
            buffer.writeln(_foldLine(
                'DTEND;TZID=${timeZone!.tzId}:${DateFormat('yyyyMMdd\'T\'HHmmss').format(endDate!)}'));
          }
        }
      } else {
        buffer.writeln(_foldLine(
            'DTSTART:${DateFormat('yyyyMMdd\'T\'HHmmss\'Z\'').format(startDate!)}'));
        if (endDate != null) {
          buffer.writeln(_foldLine(
              'DTEND:${DateFormat('yyyyMMdd\'T\'HHmmss\'Z\'').format(endDate!)}'));
        }
      }
    }

    if (recurrenceRule != null) {
      buffer.writeln(_foldLine(recurrenceRule!.toICalString()));
    }

    if (xDates != null) {
      for (final xDate in xDates!) {
        buffer.writeln(_foldLine(xDate.toICalString()));
      }
    }

    if (stamp != null) {
      buffer.writeln(_foldLine(
          'DTSTAMP:${DateFormat('yyyyMMdd\'T\'HHmmss\'Z\'').format(stamp!.toUtc())}'));
    }

    if (organizer != null) {
      buffer.writeln(_foldLine(organizer!.toICalString()));
    }

    buffer.writeln(_foldLine('UID:$uid'));

    if (attendees != null && attendees!.isNotEmpty) {
      for (final attendee in attendees!) {
        buffer.writeln(_foldLine(attendee.toICalString()));
      }
    }

    if (customProperties != null &&
        customProperties!.isNotEmpty &&
        customProperties!.containsKey('X-GOOGLE-CONFERENCE')) {
      for (final key in customProperties!.keys) {
        if (!key.contains('X-LIC')) {
          buffer.writeln(_foldLine('$key:${customProperties![key]}'));
        }
      }
    }

    if (eventClass != null) {
      buffer.writeln(_foldLine('CLASS:$eventClass'));
    }

    if (recurrenceId != null) {
      buffer.writeln(_foldLine('RECURRENCE-ID;$recurrenceId'));
    }

    if (created != null) {
      buffer.writeln(_foldLine(
          'CREATED:${DateFormat('yyyyMMdd\'T\'HHmmss\'Z\'').format(created!.toUtc())}'));
    }

    if (description != null) {
      buffer.writeln(_foldLine('DESCRIPTION:$description'));
    }

    if (lastModified != null) {
      buffer.writeln(_foldLine(
          'LAST-MODIFIED:${DateFormat('yyyyMMdd\'T\'HHmmss\'Z\'').format(lastModified!.toUtc())}'));
    }

    if (location != null) {
      buffer.writeln(_foldLine(location!.toICalString()));
    }

    if (sequence != null) {
      buffer.writeln(_foldLine('SEQUENCE:$sequence'));
    }
    if (status != null) {
      buffer.writeln(_foldLine(status!.toICalString()));
    }
    if (summary != null) {
      buffer.writeln(_foldLine('SUMMARY:$summary'));
    }
    if (transparency != null) {
      buffer.writeln(_foldLine('TRANSP:$transparency'));
    }

    if (attachments != null && attachments!.isNotEmpty) {
      for (final attachment in attachments!) {
        buffer.writeln(_foldLine(attachment.toICalString()));
      }
    }

    if (customProperties != null &&
        customProperties!.isNotEmpty &&
        !customProperties!.containsKey('X-GOOGLE-CONFERENCE')) {
      for (final key in customProperties!.keys) {
        if (!key.contains('X-LIC')) {
          buffer.writeln(_foldLine('$key:${customProperties![key]}'));
        }
      }
    }

    if (availability != null) {
      buffer.writeln(_foldLine('X-AVAILABILITY:${availability!.index}'));
    }

    if (customProperties != null && customProperties!.isNotEmpty) {
      for (final key in customProperties!.keys) {
        if (key.contains('X-LIC')) {
          buffer.writeln(_foldLine('$key:${customProperties![key]}'));
        }
      }
    }

    if (attachFiles != null && attachFiles!.isNotEmpty) {
      for (final attachFile in attachFiles!) {
        buffer.writeln(_foldLine(attachFile.toICalString()));
      }
    }

    buffer.writeln(_foldLine('END:VEVENT'));
    return buffer.toString();
  }

  factory Event.fromICalString(String icalString) {
    final lines = LineSplitter.split(icalString).toList();
    final uid = lines
        .firstWhere((line) => line.startsWith('UID:'), orElse: () => '')
        .replaceAll('UID:', '');
    final summary = lines
        .firstWhere((line) => line.startsWith('SUMMARY:'), orElse: () => '')
        .replaceAll('SUMMARY:', '');
    final description = lines
        .firstWhere((line) => line.startsWith('DESCRIPTION:'), orElse: () => '')
        .replaceAll('DESCRIPTION:', '');

    List<Attachment>? attachments;
    if (lines.any((line) => line.startsWith('ATTACH:'))) {
      attachments = [];
      for (var i = 0; i < lines.length; i++) {
        final line = lines[i];
        if (line.startsWith('ATTACH:')) {
          final attachment = Attachment.fromICalString(line);
          attachments.add(attachment);
        }
      }
    }

    List<Attendee>? attendees;
    if (lines.any((line) => line.startsWith('ATTENDEE;'))) {
      attendees = [];
      for (var i = 0; i < lines.length; i++) {
        final line = lines[i];
        if (line.startsWith('ATTENDEE;')) {
          final attendee = Attendee.fromICalString(line);
          attendees.add(attendee);
        }
      }
    }

    final eventClass = lines
        .firstWhere((line) => line.startsWith('CLASS:'), orElse: () => '')
        .replaceAll('CLASS:', '');

    final organizer = Organizer.fromICalString(
      lines.firstWhere(
        (line) => line.startsWith('ORGANIZER;'),
        orElse: () => '',
      ),
    );

    final created = lines
        .firstWhere((line) => line.startsWith('CREATED:'), orElse: () => '')
        .replaceAll('CREATED:', '');
    final lastModified = lines
        .firstWhere((line) => line.startsWith('LAST-MODIFIED:'),
            orElse: () => '')
        .replaceAll('LAST-MODIFIED:', '');
    final startDate = lines
        .firstWhere((line) => line.startsWith('DTSTART:'), orElse: () => '')
        .replaceAll('DTSTART:', '');
    final endDate = lines
        .firstWhere((line) => line.startsWith('DTEND:'), orElse: () => '')
        .replaceAll('DTEND:', '');

    RecurrenceRule? recurrenceRule;
    if (lines.any((line) => line.startsWith('RRULE:'))) {
      recurrenceRule = RecurrenceRule.fromICalString(
        lines.firstWhere((line) => line.startsWith('RRULE:')),
      );
    }

    final recurrenceId = lines
        .firstWhere((line) => line.startsWith('RECURRENCE-ID:'),
            orElse: () => '')
        .replaceAll('RECURRENCE-ID:', '');

    final sequence = lines
        .firstWhere((line) => line.startsWith('SEQUENCE:'), orElse: () => '')
        .replaceAll('SEQUENCE:', '');

    final status = lines
        .firstWhere((line) => line.startsWith('STATUS:'), orElse: () => '')
        .replaceAll('STATUS:', '');

    final transparency = lines
        .firstWhere((line) => line.startsWith('TRANSP:'), orElse: () => '')
        .replaceAll('TRANSP:', '');

    final location = lines
        .firstWhere((line) => line.startsWith('LOCATION:'), orElse: () => '')
        .replaceAll('LOCATION:', '');
    List<Attachment>? attachFiles;
    if (lines.any((line) => line.startsWith('ATTACH;'))) {
      attachFiles = [];
      for (var i = 0; i < lines.length; i++) {
        final line = lines[i];
        if (line.startsWith('ATTACH;')) {
          final attachFile = Attachment.fromICalString(line);
          attachFiles.add(attachFile);
        }
      }
    }

    var customProperties = <String, String>{};
    for (var i = 0; i < lines.length; i++) {
      final line = lines[i];
      if (line.startsWith('X-')) {
        final key = line.split(':')[0];
        final value = line.split(':')[1];
        customProperties[key] = value;
      }
    }

    return Event(
      uid: uid,
      summary: summary.isNotEmpty ? summary : null,
      description: description.isNotEmpty ? description : null,
      attachments: attachments,
      attendees: attendees,
      organizer: organizer,
      created: created.isNotEmpty ? DateTime.parse(created) : null,
      lastModified:
          lastModified.isNotEmpty ? DateTime.parse(lastModified) : null,
      startDate: startDate.isNotEmpty ? DateTime.parse(startDate) : null,
      endDate: endDate.isNotEmpty ? DateTime.parse(endDate) : null,
      recurrenceRule: recurrenceRule,
      eventClass: eventClass.isNotEmpty ? eventClass : null,
      customProperties: customProperties,
      recurrenceId: recurrenceId.isNotEmpty ? recurrenceId : null,
      sequence: sequence.isNotEmpty ? int.parse(sequence) : null,
      transparency: transparency.isNotEmpty ? transparency : null,
      location: location.isNotEmpty ? Location.fromICalString(location) : null,
      attachFiles: attachFiles,
    );
  }

  factory Event.fromICalStrings(List<String> icalStrings) {
    String? uid;
    String? summary;
    String? description;
    List<Attachment>? attachments;
    List<Attendee>? attendees;
    Organizer? organizer;
    DateTime? created;
    DateTime? lastModified;
    DateTime? startDate;
    DateTime? endDate;
    DateTime? stamp;
    int? sequence;
    String? transparency;
    EventStatus? status;
    TimeZone? timezone;
    RecurrenceRule? recurrenceRule;
    List<XDate>? exDates;
    Location? location;
    String? recurrenceId;
    String? eventClass;
    Map<String, String>? customProperties;
    List<Attachment>? attachFiles;

    List<int> indicesToSkip = [];

    for (int i = 0; i < icalStrings.length; i++) {
      if (indicesToSkip.contains(i)) {
        continue;
      }
      var icalString = icalStrings[i];
      int j = i + 1;

      while (j < icalStrings.length && icalStrings[j].startsWith(' ')) {
        String stringToAdd = icalStrings[j].substring(1);
        icalString += stringToAdd;
        indicesToSkip.add(j);
        j++;
      }

      if (icalString.startsWith('UID:')) {
        uid = icalString.split('UID:').last.trim();
      } else if (icalString.startsWith('SUMMARY:')) {
        summary = icalString.split('SUMMARY:').last.trim();
      } else if (icalString.startsWith('DESCRIPTION:')) {
        description = icalString.split('DESCRIPTION:').last.trim();
      } else if (icalString.startsWith('ATTACH:')) {
        attachments ??= [];
        attachments.add(Attachment.fromICalString(icalString));
      } else if (icalString.startsWith('ATTENDEE;')) {
        attendees ??= [];
        attendees.add(Attendee.fromICalString(icalString));
      } else if (icalString.startsWith('ORGANIZER')) {
        organizer = Organizer.fromICalString(icalString);
      } else if (icalString.startsWith('CREATED:')) {
        final createdString = icalString.split('CREATED:').last.trim();
        created = DateTime.parse(createdString);
      } else if (icalString.startsWith('LAST-MODIFIED:')) {
        final lastModifiedString =
            icalString.split('LAST-MODIFIED:').last.trim();
        lastModified = DateTime.parse(lastModifiedString);
      } else if (icalString.startsWith(
        'DTSTART:',
      )) {
        final startDateString = icalString.split('DTSTART:').last.trim();
        startDate = DateTime.parse(startDateString);
      } else if (icalString.startsWith('DTEND:')) {
        final endDateString = icalString.split('DTEND:').last.trim();
        endDate = DateTime.parse(endDateString);
      } else if (icalString.startsWith('RRULE:')) {
        recurrenceRule = RecurrenceRule.fromICalString(icalString);
      } else if (icalString.startsWith('DTSTAMP:')) {
        final stampString = icalString.split('DTSTAMP:').last.trim();
        stamp = DateTime.parse(stampString);
      } else if (icalString.startsWith('SEQUENCE:')) {
        sequence = int.parse(icalString.split('SEQUENCE:').last.trim());
      } else if (icalString.startsWith('STATUS:')) {
        String statusString = icalString.split('STATUS:').last.trim();
        status = EventStatus.fromStatusString(statusString);
      } else if (icalString.startsWith('TRANSP:')) {
        transparency = icalString.split('TRANSP:').last.trim();
      } else if (icalString.startsWith('DTSTART;')) {
        timezone = TimeZone.fromICalString(icalString.split('DTSTART;').last);
        startDate = DateTime.parse(
          icalString.split('DTSTART;TZID=').last.split(':').last.trim(),
        );
      } else if (icalString.startsWith('DTEND;')) {
        timezone = TimeZone.fromICalString(icalString.split('DTEND;').last);
        endDate = DateTime.parse(
          icalString.split('DTEND;TZID=').last.split(':').last.trim(),
        );
      } else if (icalString.startsWith('EXDATE;')) {
        exDates ??= [];
        exDates.add(XDate.fromICalString(icalString));
      } else if (icalString.startsWith('LOCATION:')) {
        location = Location.fromICalString(icalString);
      } else if (icalString.startsWith('RECURRENCE-ID;')) {
        recurrenceId = icalString.split('RECURRENCE-ID;').last.trim();
      } else if (icalString.startsWith('X-')) {
        customProperties ??= {};
        List<String> parts = icalString.split(':');
        if (parts.length > 1) {
          String propertyName = parts[0];
          String propertyValue = parts.sublist(1).join(':').trim();
          customProperties[propertyName] = propertyValue;
        }
      } else if (icalString.startsWith('CLASS:')) {
        eventClass = icalString.split('CLASS:').last.trim();
      } else if (icalString.startsWith('ATTACH;')) {
        attachFiles ??= [];
        attachFiles.add(Attachment.fromICalString(icalString));
      }
    }

    return Event(
      uid: uid ?? '',
      summary: summary,
      description: description,
      attachments: attachments,
      attendees: attendees,
      organizer: organizer,
      created: created,
      lastModified: lastModified,
      startDate: startDate,
      endDate: endDate,
      stamp: stamp,
      sequence: sequence,
      status: status,
      timeZone: timezone,
      transparency: transparency,
      recurrenceRule: recurrenceRule,
      xDates: exDates,
      location: location,
      recurrenceId: recurrenceId,
      customProperties: customProperties,
      eventClass: eventClass,
      attachFiles: attachFiles,
    );
  }

  @override
  bool operator ==(Object other) {
    if (identical(this, other)) return true;
    final collectionEquals = const DeepCollectionEquality().equals;

    return other is Event &&
        other.uid == uid &&
        other.summary == summary &&
        other.description == description &&
        collectionEquals(other.attachments, attachments) &&
        collectionEquals(other.attendees, attendees) &&
        other.organizer == organizer &&
        other.stamp == stamp &&
        other.created == created &&
        other.lastModified == lastModified &&
        other.startDate == startDate &&
        other.endDate == endDate &&
        other.sequence == sequence &&
        other.transparency == transparency &&
        other.status == status &&
        other.timeZone == timeZone &&
        other.recurrenceRule == recurrenceRule &&
        collectionEquals(other.xDates, xDates) &&
        other.location == location &&
        other.recurrenceId == recurrenceId &&
        other.eventClass == eventClass &&
        collectionEquals(other.customProperties, customProperties) &&
        collectionEquals(other.attachFiles, attachFiles) &&
        super == other;
  }

  @override
  int get hashCode {
    return uid.hashCode ^
        summary.hashCode ^
        description.hashCode ^
        attachments.hashCode ^
        attendees.hashCode ^
        organizer.hashCode ^
        stamp.hashCode ^
        created.hashCode ^
        lastModified.hashCode ^
        startDate.hashCode ^
        endDate.hashCode ^
        sequence.hashCode ^
        transparency.hashCode ^
        status.hashCode ^
        timeZone.hashCode ^
        recurrenceRule.hashCode ^
        xDates.hashCode ^
        location.hashCode ^
        recurrenceId.hashCode ^
        eventClass.hashCode ^
        customProperties.hashCode ^
        attachFiles.hashCode ^
        super.hashCode;
  }

  Map<String, dynamic> toMap() {
    return {
      'uid': uid,
      'summary': summary,
      'description': description,
      'attachments': attachments?.map((x) => x.toMap()).toList(),
      'attendees': attendees?.map((x) => x.toMap()).toList(),
      'organizer': organizer?.toMap(),
      'stamp': stamp?.millisecondsSinceEpoch,
      'created': created?.millisecondsSinceEpoch,
      'last_modified': lastModified?.millisecondsSinceEpoch,
      'start_date': startDate?.millisecondsSinceEpoch,
      'end_date': endDate?.millisecondsSinceEpoch,
      'sequence': sequence,
      'transparency': transparency,
      'status': status?.toMap(),
      'time_zone': timeZone?.toMap(),
      'recurrenceRule': recurrenceRule?.toMap(),
      'x_dates': xDates?.map((x) => x.toMap()).toList(),
      'location': location?.toMap(),
      'recurrence_id': recurrenceId,
      'event_class': eventClass,
      'custom_properties': customProperties,
      'attach_files': attachFiles?.map((x) => x.toMap()).toList(),
      'availability': availability?.index,
    };
  }

  factory Event.fromMap(Map<String, dynamic> map) {
    return Event(
      uid: map['uid'] ?? '',
      summary: map['summary'],
      description: map['description'],
      attachments: map['attachments'] != null
          ? List<Attachment>.from(
              map['attachments']?.map((x) => Attachment.fromMap(x)))
          : null,
      attendees: map['attendees'] != null
          ? List<Attendee>.from(
              map['attendees']?.map((x) => Attendee.fromMap(x)))
          : null,
      organizer:
          map['organizer'] != null ? Organizer.fromMap(map['organizer']) : null,
      stamp: map['stamp'] != null
          ? DateTime.fromMillisecondsSinceEpoch(map['stamp'])
          : null,
      created: map['created'] != null
          ? DateTime.fromMillisecondsSinceEpoch(map['created'])
          : null,
      lastModified: map['last_modified'] != null
          ? DateTime.fromMillisecondsSinceEpoch(map['last_modified'])
          : null,
      startDate: map['start_date'] != null
          ? DateTime.fromMillisecondsSinceEpoch(map['start_date'])
          : null,
      endDate: map['end_date'] != null
          ? DateTime.fromMillisecondsSinceEpoch(map['end_date'])
          : null,
      sequence: map['sequence']?.toInt(),
      transparency: map['transparency'],
      status: map['status'] != null ? EventStatus.fromMap(map['status']) : null,
      timeZone:
          map['time_zone'] != null ? TimeZone.fromMap(map['time_zone']) : null,
      recurrenceRule: map['recurrenceRule'] != null
          ? RecurrenceRule.fromMap(map['recurrenceRule'])
          : null,
      xDates: map['x_dates'] != null
          ? List<XDate>.from(map['x_dates']?.map((x) => XDate.fromMap(x)))
          : null,
      location:
          map['location'] != null ? Location.fromMap(map['location']) : null,
      recurrenceId: map['recurrence_id'],
      eventClass: map['event_class'],
      customProperties: Map<String, String>.from(map['custom_properties']),
      attachFiles: map['attach_files'] != null
          ? List<Attachment>.from(
              map['attach_files']?.map((x) => Attachment.fromMap(x)))
          : null,
    );
  }

  Map<String, dynamic> toJson() => toMap();

  factory Event.fromJson(String source) => Event.fromMap(json.decode(source));
}

enum Availability { busy, free, tentative, unavailable, unknown }
