import 'package:dart_icalendar/models/extensions/datetime_extension.dart';
import 'package:dart_icalendar/models/recurrence_model.dart';

extension RecurrenceRuleExtension on RecurrenceRule {
  DateTime? getNextDate(DateTime fromDate) {
    int? countValue = count;
    if (countValue != null && countValue < 1) {
      return null;
    }

    var startDate = fromDate;
    if (until != null && until!.isBefore(startDate)) {
      return null;
    }

    switch (frequency) {
      case Frequency.SECONDLY:
        startDate = _incrementSeconds(startDate);
        break;
      case Frequency.MINUTELY:
        startDate = _incrementMinutes(startDate);
        break;
      case Frequency.HOURLY:
        startDate = _incrementHours(startDate);
        break;
      case Frequency.DAILY:
        startDate = _incrementDays(startDate);
        break;
      case Frequency.WEEKLY:
        startDate = _incrementWeeks(startDate);
        break;
      case Frequency.MONTHLY:
        startDate = _incrementMonths(startDate);
        break;
      case Frequency.YEARLY:
        startDate = _incrementYears(startDate);
        break;
    }

    if (countValue != null) {
      countValue = countValue - 1;
    }

    if (until != null && startDate.isAfter(until!)) {
      return null;
    }

    return startDate;
  }

  DateTime _incrementSeconds(DateTime date) {
    final nextDate = date.add(Duration(seconds: interval));
    return _applyByRules(nextDate, _incrementSeconds, bySecond);
  }

  DateTime _incrementMinutes(DateTime date) {
    final nextDate = date.add(Duration(minutes: interval));
    return _applyByRules(nextDate, _incrementMinutes, byMinute);
  }

  DateTime _incrementHours(DateTime date) {
    final nextDate = date.add(Duration(hours: interval));
    return _applyByRules(nextDate, _incrementHours, byHour);
  }

  DateTime _incrementWeeks(DateTime date) {
    final nextDate = date.add(Duration(days: interval * 7));
    final byDayValues =
        byDay?.map((day) => day?.weekday.index ?? 0 + 1).toList();
    return _applyByRules(
      nextDate,
      _incrementWeeks,
      byDayValues,
      byMonthDay,
      byYearDay,
      byWeekNo,
      bySetPosition,
    );
  }

  DateTime _incrementDays(DateTime date) {
    final nextDate = date.add(Duration(days: interval));
    return _applyByRules(nextDate, _incrementDays, byMonthDay, byYearDay,
        byWeekNo, bySetPosition);
  }

  DateTime _incrementMonths(DateTime date) {
    var nextDate = DateTime(date.year, date.month + interval, date.day);
    while (nextDate.month != (date.month + interval) % 12) {
      nextDate = nextDate.subtract(Duration(days: 1));
    }
    return _applyByRules(nextDate, _incrementMonths, byMonth, byMonthDay,
        byYearDay, byWeekNo, bySetPosition);
  }

  DateTime _incrementYears(DateTime date) {
    final nextDate = DateTime(date.year + interval, date.month, date.day);
    return _applyByRules(nextDate, _incrementYears, byYearDay);
  }

  DateTime _applyByRules(
    DateTime date,
    Function(DateTime) incrementFn, [
    List<int>? byValue,
    List<int>? byMonthDay,
    List<int>? byYearDay,
    List<int>? byWeekNo,
    List<int>? bySetPosition,
  ]) {
    while (true) {
      if (byValue != null &&
          byValue.isNotEmpty &&
          !byValue.contains(date.second)) {
        date = incrementFn(date);
        continue;
      }
      if (byValue != null &&
          byValue.isNotEmpty &&
          !byValue.contains(date.minute)) {
        date = incrementFn(date);
        continue;
      }
      if (byValue != null &&
          byValue.isNotEmpty &&
          !byValue.contains(date.hour)) {
        date = incrementFn(date);
        continue;
      }
      if (byMonthDay != null &&
          byMonthDay.isNotEmpty &&
          !byMonthDay.contains(date.day)) {
        date = incrementFn(date);
        continue;
      }
      if (byYearDay != null &&
          byYearDay.isNotEmpty &&
          !byYearDay.contains(date.dayOfYear)) {
        date = incrementFn(date);
        continue;
      }
      if (byWeekNo != null &&
          byWeekNo.isNotEmpty &&
          !byWeekNo.contains(date.weekOfYear)) {
        date = incrementFn(date);
        continue;
      }
      if (bySetPosition != null && bySetPosition.isNotEmpty) {
        final monthStartDate = DateTime(date.year, date.month, 1);
        final monthStartWeekday = monthStartDate.weekday;
        final setPositions = bySetPosition
            .map((pos) => pos > 0 ? pos : monthStartDate.daysInMonth + pos + 1);
        final validDates = setPositions.map((pos) {
          if (pos > 0) {
            var dayOfWeek = (monthStartWeekday + pos - 2) % 7 + 1;
            return DateTime(
                date.year, date.month, pos + (date.weekday - dayOfWeek));
          } else {
            var dayOfWeek =
                (monthStartWeekday + pos + monthStartDate.daysInMonth) % 7 + 1;
            return DateTime(
                date.year,
                date.month,
                monthStartDate.daysInMonth +
                    pos +
                    1 +
                    (date.weekday - dayOfWeek));
          }
        });
        if (!validDates.contains(date)) {
          date = incrementFn(date);
          continue;
        }
      }
      break;
    }
    return date;
  }
}
