import 'dart:convert';

import 'package:collection/collection.dart';
import 'package:intl/intl.dart';

import 'package:dart_icalendar/models/attachment_model.dart';
import 'package:dart_icalendar/models/person_model.dart';
import 'package:dart_icalendar/models/recurrence_model.dart';

class Journal {
  String uid;
  String? summary;
  String? description;
  List<Attachment>? attachments;
  List<Attendee>? attendees;
  Organizer? organizer;
  DateTime? created;
  DateTime? lastModified;
  DateTime? startDate;
  DateTime? endDate;
  RecurrenceRule? recurrenceRule;

  Journal({
    required this.uid,
    this.summary,
    this.description,
    this.attachments,
    this.attendees,
    this.organizer,
    this.created,
    this.lastModified,
    this.startDate,
    this.endDate,
    this.recurrenceRule,
  });

  String toICalString() {
    final buffer = StringBuffer();
    buffer.writeln('BEGIN:VJOURNAL');
    buffer.writeln('UID:$uid');
    if (summary != null) {
      buffer.writeln('SUMMARY:$summary');
    }
    if (description != null) {
      buffer.writeln('DESCRIPTION:$description');
    }
    if (attachments != null && attachments!.isNotEmpty) {
      for (final attachment in attachments!) {
        buffer.writeln(attachment.toICalString());
      }
    }
    if (attendees != null && attendees!.isNotEmpty) {
      for (final attendee in attendees!) {
        buffer.writeln(attendee.toICalString());
      }
    }
    if (organizer != null) {
      buffer.writeln(organizer!.toICalString());
    }
    if (created != null) {
      buffer.writeln(
          'CREATED:${DateFormat('yyyyMMdd\'T\'HHmmss\'Z\'').format(created!.toUtc())}');
    }
    if (lastModified != null) {
      buffer.writeln(
          'LAST-MODIFIED:${DateFormat('yyyyMMdd\'T\'HHmmss\'Z\'').format(lastModified!.toUtc())}');
    }
    if (startDate != null) {
      buffer.writeln(
          'DTSTART:${DateFormat('yyyyMMdd\'T\'HHmmss\'Z\'').format(startDate!.toUtc())}');
    }
    if (endDate != null) {
      buffer.writeln(
          'DTEND:${DateFormat('yyyyMMdd\'T\'HHmmss\'Z\'').format(endDate!.toUtc())}');
    }
    if (recurrenceRule != null) {
      buffer.writeln(recurrenceRule!.toICalString());
    }
    buffer.writeln('END:VJOURNAL');
    return buffer.toString();
  }

  Journal copyWith({
    String? uid,
    String? summary,
    String? description,
    List<Attachment>? attachments,
    List<Attendee>? attendees,
    Organizer? organizer,
    DateTime? created,
    DateTime? lastModified,
    DateTime? startDate,
    DateTime? endDate,
    RecurrenceRule? recurrenceRule,
  }) {
    return Journal(
      uid: uid ?? this.uid,
      summary: summary ?? this.summary,
      description: description ?? this.description,
      attachments: attachments ?? this.attachments,
      attendees: attendees ?? this.attendees,
      organizer: organizer ?? this.organizer,
      created: created ?? this.created,
      lastModified: lastModified ?? this.lastModified,
      startDate: startDate ?? this.startDate,
      endDate: endDate ?? this.endDate,
      recurrenceRule: recurrenceRule ?? this.recurrenceRule,
    );
  }

  Map<String, dynamic> toMap() {
    return {
      'uid': uid,
      'summary': summary,
      'description': description,
      'attachments': attachments?.map((x) => x.toMap()).toList(),
      'attendees': attendees?.map((x) => x.toMap()).toList(),
      'organizer': organizer?.toMap(),
      'created': created?.millisecondsSinceEpoch,
      'lastModified': lastModified?.millisecondsSinceEpoch,
      'start_date': startDate?.millisecondsSinceEpoch,
      'end_date': endDate?.millisecondsSinceEpoch,
      'recurrenceRule': recurrenceRule?.toMap(),
    };
  }

  factory Journal.fromMap(Map<String, dynamic> map) {
    return Journal(
      uid: map['uid'] ?? '',
      summary: map['summary'],
      description: map['description'],
      attachments: map['attachments'] != null
          ? List<Attachment>.from(
              map['attachments']?.map((x) => Attachment.fromMap(x)))
          : null,
      attendees: map['attendees'] != null
          ? List<Attendee>.from(
              map['attendees']?.map((x) => Attendee.fromMap(x)))
          : null,
      organizer:
          map['organizer'] != null ? Organizer.fromMap(map['organizer']) : null,
      created: map['created'] != null
          ? DateTime.fromMillisecondsSinceEpoch(map['created'])
          : null,
      lastModified: map['last_modified'] != null
          ? DateTime.fromMillisecondsSinceEpoch(map['last_modified'])
          : null,
      startDate: map['start_date'] != null
          ? DateTime.fromMillisecondsSinceEpoch(map['start_date'])
          : null,
      endDate: map['end_date'] != null
          ? DateTime.fromMillisecondsSinceEpoch(map['endDate'])
          : null,
      recurrenceRule: map['recurrence_rule'] != null
          ? RecurrenceRule.fromMap(map['recurrence_rule'])
          : null,
    );
  }

  String toJson() => json.encode(toMap());

  factory Journal.fromJson(String source) =>
      Journal.fromMap(json.decode(source));

  @override
  String toString() {
    return 'Journal(uid: $uid, summary: $summary, description: $description, attachments: $attachments, attendees: $attendees, organizer: $organizer, created: $created, lastModified: $lastModified, startDate: $startDate, endDate: $endDate, recurrenceRule: $recurrenceRule)';
  }

  @override
  bool operator ==(Object other) {
    if (identical(this, other)) return true;
    final listEquals = const DeepCollectionEquality().equals;

    return other is Journal &&
        other.uid == uid &&
        other.summary == summary &&
        other.description == description &&
        listEquals(other.attachments, attachments) &&
        listEquals(other.attendees, attendees) &&
        other.organizer == organizer &&
        other.created == created &&
        other.lastModified == lastModified &&
        other.startDate == startDate &&
        other.endDate == endDate &&
        other.recurrenceRule == recurrenceRule;
  }

  @override
  int get hashCode {
    return uid.hashCode ^
        summary.hashCode ^
        description.hashCode ^
        attachments.hashCode ^
        attendees.hashCode ^
        organizer.hashCode ^
        created.hashCode ^
        lastModified.hashCode ^
        startDate.hashCode ^
        endDate.hashCode ^
        recurrenceRule.hashCode;
  }
}
