import 'dart:convert';

class Person {
  String? name;
  String? email;
  String? directoryEntry;

  Person({
    required this.name,
    required this.email,
    this.directoryEntry,
  });

  String toICalString() {
    final buffer = StringBuffer();
    buffer.writeln('CN:$name');
    buffer.writeln('EMAIL:$email');
    if (directoryEntry != null) {
      buffer.writeln('DIR;$directoryEntry');
    }
    return buffer.toString();
  }

  Person copyWith({
    String? name,
    String? email,
    String? directoryEntry,
  }) {
    return Person(
      name: name ?? this.name,
      email: email ?? this.email,
      directoryEntry: directoryEntry ?? this.directoryEntry,
    );
  }

  Map<String, dynamic> toMap() {
    return {
      'name': name,
      'email': email,
      'directory_entry': directoryEntry,
    };
  }

  factory Person.fromMap(Map<String, dynamic> map) {
    return Person(
      name: map['name'] ?? '',
      email: map['email'] ?? '',
      directoryEntry: map['directory_entry'],
    );
  }

  String toJson() => json.encode(toMap());

  factory Person.fromJson(String source) => Person.fromMap(json.decode(source));

  @override
  String toString() =>
      'Person(name: $name, email: $email, directoryEntry: $directoryEntry)';

  @override
  bool operator ==(Object other) {
    if (identical(this, other)) return true;

    return other is Person &&
        other.name == name &&
        other.email == email &&
        other.directoryEntry == directoryEntry;
  }

  @override
  int get hashCode => name.hashCode ^ email.hashCode ^ directoryEntry.hashCode;
}

class Attendee extends Person {
  String? role;
  String? partstat;
  bool? rsvp;
  String? delegate;
  String? cutype;
  int? numGuests;
  String? mailTo;

  Attendee({
    required String name,
    required String email,
    this.role,
    this.partstat,
    this.rsvp,
    this.delegate,
    this.cutype,
    this.numGuests,
    this.mailTo,
    String? directoryEntry,
  }) : super(name: name, email: email, directoryEntry: directoryEntry);

  @override
  String toICalString() {
    final buffer = StringBuffer();
    buffer.write(
        'ATTENDEE;CUTYPE=INDIVIDUAL;ROLE=${role ?? 'REQ-PARTICIPANT'};PARTSTAT=${partstat ?? 'NEEDS-ACTION'}');
    if (name != null && name!.isNotEmpty) {
      buffer.write(';CN=$name');
    }
    if (numGuests != null) {
      buffer.write(';X-NUM-GUESTS=$numGuests');
    }
    if (rsvp != null) {
      buffer.write(';RSVP=$rsvp');
    }
    if (mailTo != null) {
      buffer.write(':mailto:$mailTo');
    } else {
      buffer.write(':');
    }
    buffer.writeln();
    if (delegate != null) {
      buffer.writeln('DELEGATED-TO=$delegate');
    }
    return buffer.toString().trim();
  }

  static String removeNonPrintableChars(String input) {
    RegExp exp = RegExp(r'[^\x20-\x7E]');
    return input.replaceAll(exp, '');
  }

  factory Attendee.fromICalString(String icalString) {
    String name = '';
    String email = '';
    String? cutype;
    String? rsvp;
    int? numGuests;
    String? mailTo;
    var fields = icalString.split(';');

    for (int i = 0; i < fields.length; i++) {
      fields[i] = removeNonPrintableChars(fields[i]);
    }

    for (final field in fields) {
      if (field.startsWith('CN=')) {
        final cnMatch = RegExp(r'(^|\s)CN=([^;]+)').firstMatch(field);
        if (cnMatch != null) {
          name = cnMatch.group(2) ?? '';
        }
      } else if (field.startsWith('EMAIL=')) {
        email = field.substring(6);
      } else if (field.startsWith('CUTYPE=')) {
        cutype = field.substring(7);
      } else if (field.startsWith('RSVP=')) {
        rsvp = field.substring(5);
      } else if (field.trim().startsWith('X-NUM-GUESTS=')) {
        numGuests = int.tryParse(field.substring(13).split(':')[0]);
        mailTo = field.substring(13).split(':').last;
      }
    }
    final role = fields
        .firstWhere((field) => field.startsWith('ROLE='),
            orElse: () => 'ROLE=REQ-PARTICIPANT')
        .substring(5);
    final partstatField = fields.firstWhere(
        (field) => field.startsWith('PARTSTAT='),
        orElse: () => 'PARTSTAT=NEEDS-ACTION');
    final partstat = partstatField.substring(9);

    String? delegate = fields.firstWhere(
        (field) => field.startsWith('DELEGATED-TO='),
        orElse: () => 'null');
    if (delegate != 'null') {
      delegate = delegate.substring(13);
    } else {
      delegate = null;
    }

    return Attendee(
      name: name,
      email: email,
      role: role,
      partstat: partstat,
      rsvp: rsvp != null ? rsvp == 'TRUE' : null,
      delegate: delegate,
      cutype: cutype,
      numGuests: numGuests,
      mailTo: mailTo,
    );
  }

  @override
  bool operator ==(Object other) {
    if (identical(this, other)) return true;

    return other is Attendee &&
        other.role == role &&
        other.partstat == partstat &&
        other.rsvp == rsvp &&
        other.delegate == delegate;
  }

  @override
  int get hashCode {
    return role.hashCode ^
        partstat.hashCode ^
        rsvp.hashCode ^
        delegate.hashCode;
  }

  @override
  Map<String, dynamic> toMap() {
    return {
      'role': role,
      'partstat': partstat,
      'rsvp': rsvp,
      'delegate': delegate,
      'cutype': cutype,
      'num_guests': numGuests,
      'mail_to': mailTo,
    };
  }

  factory Attendee.fromMap(Map<String, dynamic> map) {
    return Attendee(
      name: map['name'],
      email: map['email'],
      role: map['role'],
      partstat: map['partstat'],
      rsvp: map['rsvp'],
      delegate: map['delegate'],
      cutype: map['cutype'],
      numGuests: map['num_guests']?.toInt(),
      mailTo: map['mail_to'],
    );
  }

  factory Attendee.fromJson(String source) =>
      Attendee.fromMap(json.decode(source));
}

class Contact extends Person {
  String? phone;
  String? address;

  Contact(
      {required String name,
      required String email,
      this.phone,
      this.address,
      String? directoryEntry})
      : super(name: name, email: email, directoryEntry: directoryEntry);

  @override
  String toICalString() {
    final buffer = StringBuffer();
    buffer.writeln('CONTACT:${super.toICalString().trim()}');
    if (phone != null) {
      buffer.writeln('TEL:$phone');
    }
    if (address != null) {
      buffer.writeln('ADR:$address');
    }
    return buffer.toString();
  }

  factory Contact.fromICalString(String icalString) {
    final name =
        icalString.replaceAll('CONTACT', '').split(';').first.split(':')[1];
    final email =
        icalString.replaceAll('CONTACT', '').split(';').last.split(':')[1];
    final phone =
        icalString.contains('TEL=') ? icalString.split('TEL=').last : null;
    final address =
        icalString.contains('ADR=') ? icalString.split('ADR=').last : null;

    return Contact(
      name: name,
      email: email,
      phone: phone,
      address: address,
    );
  }
}

class Organizer extends Person {
  String? sentBy;

  Organizer(
      {required String name,
      required String email,
      this.sentBy,
      String? directoryEntry})
      : super(name: name, email: email, directoryEntry: directoryEntry);

  @override
  String toICalString() {
    if (name == null || name!.isEmpty) {
      return 'ORGANIZER:mailto:$email\r';
    }
    return 'ORGANIZER;CN=$name:mailto:$email\r';
  }

  factory Organizer.fromICalString(String icalString) {
    final separatorIndex = icalString.indexOf(':');
    if (separatorIndex < 0) {
      throw FormatException('Invalid iCalendar string for Organizer');
    }

    final colonIndex = icalString.indexOf(':');
    final semiColonIndex = icalString.indexOf(';');
    final delimiterIndex = (semiColonIndex >= 0 && semiColonIndex < colonIndex)
        ? semiColonIndex
        : colonIndex;

    if (delimiterIndex < 0) {
      throw FormatException('Invalid iCalendar string for Organizer');
    }

    final mailtoIndex = icalString.indexOf('mailto:', delimiterIndex);
    if (mailtoIndex < 0) {
      throw FormatException('Invalid iCalendar string for Organizer');
    }

    final nameAndEmail = icalString.substring(delimiterIndex + 1, mailtoIndex);
    final parts = nameAndEmail.split('=');
    String name = (parts.length == 2 && parts[0] == 'CN')
        ? parts[1].replaceAll('"', '')
        : '';
    String email = icalString.substring(mailtoIndex + 'mailto:'.length).trim();

    name = name.replaceAll(RegExp(r'[:;]$'), '');
    email = email.replaceAll(RegExp(r'[:;]$'), '');

    return Organizer(
      name: name,
      email: email,
    );
  }

  @override
  bool operator ==(Object other) {
    if (identical(this, other)) return true;

    return other is Organizer && other.sentBy == sentBy;
  }

  @override
  int get hashCode => sentBy.hashCode;

  @override
  Map<String, dynamic> toMap() {
    return {
      'sentBy': sentBy,
    };
  }

  factory Organizer.fromMap(Map<String, dynamic> map) {
    return Organizer(
      name: map['name'],
      email: map['email'],
      sentBy: map['sentBy'],
      directoryEntry: map['directoryEntry'],
    );
  }

  factory Organizer.fromJson(String source) =>
      Organizer.fromMap(json.decode(source));
}
