import 'dart:convert';

import 'package:collection/collection.dart';

class TimeZone {
  final String tzId;
  final List<TimeZoneTransition> transitions;

  TimeZone({
    required this.tzId,
    required this.transitions,
  });

  bool get isUtc => tzId == 'UTC';

  static TimeZone get utc => TimeZone(tzId: 'UTC', transitions: []);

  factory TimeZone.fromICalString(String icalString) {
    final lines = icalString.split('\n');
    String? tzId;
    final transitions = <TimeZoneTransition>[];
    tzId = lines.first.split('=')[1].split(':')[0];
    return TimeZone(
      tzId: tzId,
      transitions: transitions,
    );
  }
  static DateTime timeZoneConvert(DateTime dateTime, String tzid) {
    switch (tzid) {
      case 'Etc/GMT+12':
        return dateTime.add(Duration(hours: -12));
      case 'Pacific/Midway':
        return dateTime.add(Duration(hours: -11));
      case 'Pacific/Honolulu':
        return dateTime.add(Duration(hours: -10));
      case 'America/Juneau':
        return dateTime.add(Duration(hours: -9));
      case 'America/Los_Angeles':
        return dateTime.add(Duration(hours: -8));
      case 'America/Phoenix':
        return dateTime.add(Duration(hours: -7));
      case 'America/Denver':
        return dateTime.add(Duration(hours: -7));
      case 'America/Chicago':
        return dateTime.add(Duration(hours: -6));
      case 'America/New_York':
        return dateTime.add(Duration(hours: -5));
      case 'America/Argentina/Buenos_Aires':
        return dateTime.add(Duration(hours: -3));
      case 'America/Sao_Paulo':
        return dateTime.add(Duration(hours: -3));
      case 'Atlantic/South_Georgia':
        return dateTime.add(Duration(hours: -2));
      case 'Atlantic/Azores':
        return dateTime.add(Duration(hours: -1));
      case 'Europe/London':
        return dateTime.add(Duration(hours: 0));
      case 'Europe/Paris':
        return dateTime.add(Duration(hours: 1));
      case 'Europe/Moscow':
        return dateTime.add(Duration(hours: 3));
      case 'Asia/Dubai':
        return dateTime.add(Duration(hours: 4));
      case 'Asia/Kolkata':
        return dateTime.add(Duration(hours: 5, minutes: 30));
      case 'Asia/Hong_Kong':
        return dateTime.add(Duration(hours: 8));
      case 'Asia/Tokyo':
        return dateTime.add(Duration(hours: 9));
      case 'Australia/Sydney':
        return dateTime.add(Duration(hours: 10));
      case 'Pacific/Fiji':
        return dateTime.add(Duration(hours: 12));
      case 'Pacific/Auckland':
        return dateTime.add(Duration(hours: 12));
      case 'America/Toronto':
        return dateTime.add(Duration(hours: -5));
      case 'DATE':
        return dateTime;

      default:
        throw Exception('Invalid Time Zone');
    }
  }

  String toICalString() {
    final buffer = StringBuffer();
    buffer.writeln('BEGIN:VTIMEZONE');
    buffer.writeln('TZID:$tzId');
    for (final transition in transitions) {
      buffer.writeln(transition.toICalString());
    }
    buffer.writeln('END:VTIMEZONE');
    return buffer.toString();
  }

  TimeZone copyWith({
    String? tzId,
    List<TimeZoneTransition>? transitions,
  }) {
    return TimeZone(
      tzId: tzId ?? this.tzId,
      transitions: transitions ?? this.transitions,
    );
  }

  Map<String, dynamic> toMap() {
    return {
      'tz_id': tzId,
      'transitions': transitions.map((x) => x.toMap()).toList(),
    };
  }

  factory TimeZone.fromMap(Map<String, dynamic> map) {
    return TimeZone(
      tzId: map['tz_id'] ?? '',
      transitions: List<TimeZoneTransition>.from(
          map['transitions']?.map((x) => TimeZoneTransition.fromMap(x))),
    );
  }

  String toJson() => json.encode(toMap());

  factory TimeZone.fromJson(String source) =>
      TimeZone.fromMap(json.decode(source));

  @override
  String toString() => 'TimeZone(tzId: $tzId, transitions: $transitions)';

  @override
  bool operator ==(Object other) {
    if (identical(this, other)) return true;
    final listEquals = const DeepCollectionEquality().equals;

    return other is TimeZone &&
        other.tzId == tzId &&
        listEquals(other.transitions, transitions);
  }

  @override
  int get hashCode => tzId.hashCode ^ transitions.hashCode;
}

class TimeZoneTransition {
  final bool isDaylight;
  final DateTime dateTime;
  final String timeZoneOffset;

  TimeZoneTransition(
      {required this.isDaylight,
      required this.dateTime,
      required this.timeZoneOffset});

  factory TimeZoneTransition.fromICalString(String icalString) {
    final lines = icalString.split('\n');
    bool? isDaylight;
    DateTime? dateTime;
    String? timeZoneOffset;
    for (final line in lines) {
      if (line.startsWith('DTSTART:')) {
        dateTime = DateTime.parse(line.substring(8));
      } else if (line.startsWith('TZOFFSETTO:')) {
        timeZoneOffset = line.substring(12);
      } else if (line.startsWith('TZOFFSETFROM:')) {
      } else if (line.startsWith('RRULE:')) {
      } else if (line.startsWith('BEGIN:')) {
        isDaylight = line.endsWith(':DAYLIGHT');
      }
    }
    if (isDaylight == null || dateTime == null || timeZoneOffset == null) {
      throw FormatException('Invalid iCalendar string: $icalString');
    }
    return TimeZoneTransition(
        isDaylight: isDaylight,
        dateTime: dateTime,
        timeZoneOffset: timeZoneOffset);
  }

  String toICalString() {
    final buffer = StringBuffer();
    buffer.writeln('BEGIN:${isDaylight ? 'DAYLIGHT' : 'STANDARD'}');
    buffer.writeln('DTSTART:${_formatDateTime(dateTime)}');
    buffer.writeln('TZOFFSETTO:$timeZoneOffset');
    buffer.writeln('END:${isDaylight ? 'DAYLIGHT' : 'STANDARD'}');
    return buffer.toString();
  }

  String _formatDateTime(DateTime dateTime) {
    return '${dateTime.year.toString().padLeft(4, '0')}${dateTime.month.toString().padLeft(2, '0')}${dateTime.day.toString().padLeft(2, '0')}T${dateTime.hour.toString().padLeft(2, '0')}${dateTime.minute.toString().padLeft(2, '0')}00';
  }

  @override
  bool operator ==(Object other) {
    if (identical(this, other)) return true;

    return other is TimeZoneTransition &&
        other.isDaylight == isDaylight &&
        other.dateTime == dateTime &&
        other.timeZoneOffset == timeZoneOffset;
  }

  @override
  int get hashCode =>
      isDaylight.hashCode ^ dateTime.hashCode ^ timeZoneOffset.hashCode;

  Map<String, dynamic> toMap() {
    return {
      'isDaylight': isDaylight,
      'dateTime': dateTime.millisecondsSinceEpoch,
      'timeZoneOffset': timeZoneOffset,
    };
  }

  factory TimeZoneTransition.fromMap(Map<String, dynamic> map) {
    return TimeZoneTransition(
      isDaylight: map['isDaylight'] ?? false,
      dateTime: DateTime.fromMillisecondsSinceEpoch(map['dateTime'] ?? 0),
      timeZoneOffset: map['timeZoneOffset'] ?? '',
    );
  }

  String toJson() => json.encode(toMap());

  factory TimeZoneTransition.fromJson(String source) =>
      TimeZoneTransition.fromMap(json.decode(source));
}
